<?php
/**
* mySocketServer - adapted from patServer created by Stephan Schmidt <schst@php-tools.net>
* PHP socket server abstract class
* Events must be defined in children class:
*   * onStart
*   * onConnect
*   * onConnectionRefused
*   * onClose
*   * onShutdown
*   * onReceiveData
*
* @version  0.1
* @author   Bartabat
* @package  mySocketServer
*/

abstract class mySocketServer
{
  protected
    $port,
    $domain,

    $maxClient = -1,
    $listClient = array(),
    $clientInfo = array(),
    $totalClient = 0,
    $maxQueue = 500,

    $readBufferSize = 512,
    $readEndCharacter = "\n",

    $lifetime = NULL,

    $debug = true,
    $debugMode = 'text',
    $debugDest = 'chat_log.txt';


  public function __construct($domain='localhost', $port=10000)
  {
    $this->domain = $domain;
    $this->port = $port;
  }

  public function start()
  {
    if(!$this->initSocket = @socket_create(AF_INET, SOCK_STREAM, 0))
    {
      die("mySocketServer: Could not create socket.");
    }

    //  adress may be reused
    socket_set_option($this->initSocket, SOL_SOCKET, SO_REUSEADDR, 1);

    //  bind the socket
    if(!@socket_bind($this->initSocket, $this->domain, $this->port))
    {
      @socket_close($this->initSocket);
      die("mySocketServer: Could not bind socket to ".$this->domain." on port ".$this->port." (".$this->getLastSocketError($this->initSocket).").");
    }

    //  listen on selected port
    if(!@socket_listen( $this->initSocket, $this->maxQueue ))
    {
      die("mySocketServer: Could not listen (".$this->getLastSocketError($this->initSocket).").");
    }

    register_shutdown_function(array($this, "shutdown"));

    $this->log("Listening on port ".$this->port.". Server started at ".date("H:i:s", time()));
    $this->onStart();
    $this->run();
  }

  private function run()
  {
    while(true)
    {
      $this->onRunCycle();

      $readSocket = array();
      $writeSocket = array();
      $exceptSocket = array();

      array_push($readSocket, $this->initSocket);

      //  fetch all clients that are awaiting connections
      $nbClient = count($this->listClient);
      for($i=0; $i<$nbClient; $i++)
      {
        if(isset($this->listClient[$i]))
        {
          array_push($readSocket, $this->listClient[$i]);
        }
      }

      $this->log("---------------------");
      $this->log("BEFORE socket_select");

      //  block and wait for data or new connection - $ready reçoit le nombre de socket modifié, donc à traiter
      if(($ready = @socket_select($readSocket, $writeSocket, $exceptSocket, $this->lifetime)) === false)
      {
        $this->log("socket_select failed");
        $this->shutdown();
      }

      $this->log("---------------------");
      $this->log("AFTER socket_select");

      //  check for new connection
      if(in_array($this->initSocket, $readSocket))
      {
        $newClient = $this->acceptConnection($this->initSocket);

        //  check for maximum amount of connections
        if($this->maxClient > 0 && $this->totalClient > $this->maxClient)
        {
          $this->log("Too many connections.");
          $this->onConnectionRefused($newClient);
          $this->closeConnection($newClient);
        }

        if(--$ready <= 0)
        {
          continue;
        }
      }

      //  check all clients for incoming data
      $nbClient = count($this->listClient);
      for($i=0; $i<$nbClient; $i++)
      {
        if(!isset($this->listClient[$i]))
        {
          continue;
        }

        if(in_array($this->listClient[$i], $readSocket))
        {
          $data = $this->readFromSocket($i);

          //  empty data => connection was closed
          if(!$data)
          {
            $this->log("Connection closed by peer");
            $this->closeConnection($i);
          }
          else
          {
            $this->log("Received ".trim($data)." from ".$i);

            $this->onReceiveData($i, $data);
          }
        }
      }
    }
  }

  public function shutdown()
  {
    $this->onShutdown();

    $nbClient = count($this->listClient);
    for($i=0; $i<$nbClient; $i++)
    {
      $this->closeConnection($i);
    }

    @socket_close($this->initSocket);

    $this->log("Shutdown server.");
    exit;
  }

  protected function sendData($clientId, $data, $debugData = true)
  {
    if(!isset($this->listClient[$clientId]) || $this->listClient[$clientId] == NULL)
    {
      return false;
    }

    if($debugData)
    {
      $this->log("sending: \"" . $data . "\" to: ".$clientId);
    }

    if(!@socket_write($this->listClient[$clientId], $data))
    {
      $this->sendDebugMessage("Could not write '".$data."' client ".$clientId." (".$this->getLastSocketError($this->listClient[$clientId]).").");
    }
  }

  protected function broadcastData($data, $exclude=array())
  {
    if(!is_array($exclude))
    {
      $exclude = array($exclude);
    }

    $this->log('broadcast exlude: '.$exclude);

    $nbClient = count($this->listClient);
    for($i=0; $i<$nbClient; $i++)
    {
      if(isset($this->listClient[$i]) && $this->listClient[$i] != NULL && !in_array($i, $exclude))
      {
        if(!@socket_write($this->listClient[$i], $data))
        {
          $this->log( "Could not write '".$data."' client ".$i." (".$this->getLastSocketError($this->listClient[$i]).").");
        }
      }
    }
  }

  private function acceptConnection(&$socket)
  {
    $nbClient = count($this->listClient);

    for($i=0; $i<=$nbClient; $i++)
    {
      if(!isset($this->listClient[$i]) || $this->listClient[$i] == NULL)
      {
        $this->listClient[$i] = socket_accept($socket);
        socket_set_option($this->listClient[$i], SOL_SOCKET, SO_REUSEADDR, 1);

        $peer_host = "";
        $peer_port = "";
        socket_getpeername( $this->listClient[$i], $peer_host, $peer_port);
        $this->clientInfo[$i] = array(
          "host" => $peer_host,
          "port" => $peer_port,
          "connectOn" => time()
        );

        $this->totalClient++;
        $this->log("New connection (".$i.") from ".$peer_host." on port ".$peer_port);
        $this->onConnect($i);

        return $i;
      }
    }
  }

  protected function closeConnection($id)
  {
    if(!isset($this->listClient[$id]))
    {
      return false;
    }

    $this->onCloseConnection($id);

    $this->log("Closed connection (".$id.") from ".$this->clientInfo[$id]["host"]." on port ".$this->clientInfo[$id]["port"]);

    @socket_close($this->listClient[$id]);
    $this->listClient[$id] = NULL;
    unset($this->clientInfo[$id]);

    $this->totalClient--;
  }

  private function readFromSocket($id)
  {
    //  start with empty string
    $data = "";

    //  read data from socket
    while($buffer = @socket_read($this->listClient[$id], $this->readBufferSize))
    {
      $data .= $buffer;

      $endString = substr($buffer, - strlen($this->readEndCharacter));
      if($endString == $this->readEndCharacter)
      {
        break;
      }
    }

    if($buffer === false)
    {
      $this->log("Could not read from client ".$id." (".$this->getLastSocketError($this->listClient[$id]).").");
    }

    return $data;
  }

  protected function getLastSocketError(&$socket)
  {
    $lastError  = socket_last_error($socket);
    return "msg: ".socket_strerror($lastError)." / Code: ".$lastError;
  }

  protected function log($msg)
  {
    if(!$this->debug)
    {
      return false;
    }

    $msg = date('Y-m-d H:i:s').' mySocketServer: '.$msg;

    switch($this->debugMode)
    {
      case 'text':
        $msg = $msg."\n";
        break;
      case 'html':
        $msg = htmlspecialchars($msg)."<br />\n";
        break;
    }

    if($this->debugDest == 'stdout' || empty($this->debugDest))
    {
      echo $msg;
      flush();
      ob_flush();
      return true;
    }

    error_log($msg, 3, $this->debugDest);
    return true;
  }

  abstract protected function onStart();
  abstract protected function onRunCycle();
  abstract protected function onShutdown();
  abstract protected function onReceiveData($id, $data);
  abstract protected function onConnect($id);
  abstract protected function onConnectionRefused($id);
  abstract protected function onCloseConnection($id);
}
